
import * as fs from 'fs-extra'
import * as path from 'path'
import { IContentType } from './model'
import * as util from 'util'

const { diff } = require('json-diff')
const { colorize } = require('json-diff/lib/colorize')

export interface IArgs {
  from: string,
  to: string,
  outDir: string
}

export default async function Run(args: IArgs) {

  const [from, to] = (await Promise.all([
    fs.readFile(args.from),
    fs.readFile(args.to)
  ])).map(b => JSON.parse(b.toString()))

  const fromTypes = indexById(from.contentTypes)
  const toTypes = indexById(to.contentTypes)

  const fileName = `${new Date().toISOString().replace(/[^\d]/g, '').substring(0, 14)}_generated_from_diff.ts`
  const outputStream = fs.createWriteStream(path.join(args.outDir, fileName))
  const write = asyncWrite(outputStream)

  await write(`import Migration from 'contentful-migration-cli'

// Generated by contentful-schema-diff
// from ${args.from}
// to   ${args.to}
export = function (migration: Migration) {
`)

  const promises = Object.keys(toTypes).map(async (id) => {
    let chunk: string
    if (fromTypes[id]) {
      chunk = await writeModify(fromTypes[id], toTypes[id])
    } else {
      chunk = await writeCreate(toTypes[id])
    }

    if (chunk && chunk.length > 0) {
      await write(chunk)
    }
  })
  promises.push(...Object.keys(fromTypes).map(async (id) => {
    if (toTypes[id]) {
      // handled above in 'writeModify'
      return
    }
    const chunk = await writeDelete(id)
    await write(chunk)
  }))

  await Promise.all(promises)

  await write(`
}
`)

  outputStream.close();
}

function indexById(types: IContentType[]): { [id: string]: IContentType } {
  const ret: any = {}
  types.forEach(type => {
    ret[type.sys.id] = type
  })
  return ret
}

function asyncWrite(stream: fs.WriteStream): (chunk: any) => Promise<void> {
  let draining = true

  function doWrite(chunk: any) {
    return new Promise<void>((resolve, reject) => {
      if (draining) {
        draining = stream.write(chunk, (err: any) => {
          if(err) {
            reject(err)
          } else {
            resolve()
          }
        })
      } else {
        stream.on('drain', () => {
          // await recursive
          doWrite(chunk)
            .then(resolve)
            .catch(reject)
        })
      }
    })
  }

  return doWrite
}

async function writeDelete(id: string): Promise<string> {
  return `
  migration.deleteContentType('${id}')
`
}

async function writeModify(from: IContentType, to: IContentType): Promise<string> {
  const difference = diff(from.fields, to.fields)
  if (!difference || difference.length == 0) {
    return
  }

  const v = camelCase(from.sys.id)
  const typeDef = Object.assign({}, to)
  delete(typeDef.fields)
  delete(typeDef.sys)

  let str = `
  var ${v} = migration.editContentType('${from.sys.id}', ${dump(typeDef)})
`
    str += `
  /* TODO: automatically generate edits from this diff
${colorize(difference, { color: false } )} */
  `

  return str
}

async function writeCreate(newType: IContentType): Promise<string> {
  const v = camelCase(newType.sys.id)
  const typeDef = Object.assign({}, newType)
  delete(typeDef.fields)
  delete(typeDef.sys)

  let str = `
  var ${v} = migration.createContentType('${newType.sys.id}', ${dump(typeDef)})
`
  newType.fields.forEach(field => {
    const fieldDef = Object.assign({}, field)
    delete(fieldDef.id)

    str += `
  ${v}.createField('${field.id}', ${dump(fieldDef)})
`
  })
  
  return str;
}

function dump(obj: any): string {
  return util.inspect(obj, {
    depth: null,
    maxArrayLength: null,
    breakLength: 0
  })
}

function camelCase(input: string) { 
  return input.toLowerCase().replace(/-(.)/g, function(match, group1) {
      return group1.toUpperCase();
  });
}